==> podwiki-0.8.0.tgz: podwiki-0.8.0/lib/PodWiki/Auth.pm <==
 #
 #   $Id: Auth.pm,v 1.5 2004/05/24 19:11:58 zarahg Exp $
 #
 #   This  file  is  part of the  PodWiki  web-based authoring tool.
 #
 #   By  accessing  this software,  PodWiki,  you are  duly informed
 #   of and  agree to be  bound  by the  conditions  described below
 #   in this notice:
 #
 #   This software product, PodWiki,  is developed by  Thomas Linden
 #   and   copyrighted  (C) 2003-2004   by  Thomas Linden,  with all
 #   rights reserved.
 #
 #   There is  no charge for  PodWiki software. You can redistribute
 #   it and/or modify it under  the terms of the  GNU General Public
 #   License, which is incorporated by reference herein.
 #
 #   PodWiki is distributed WITHOUT ANY WARRANTY,IMPLIED OR EXPRESS,
 #   OF  MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE or that
 #   the  use of it will not infringe on any third party's intellec-
 #   tual property rights.
 #
 #   You  should  have  received a  copy  of  the GNU General Public
 #   License along with PodWiki.  Copies  can also be obtained from:
 #
 #     http://www.gnu.org/copyleft/gpl.html
 #
 #   or by writing to:
 #
 #     Free Software Foundation, Inc.
 #     59 Temple Place, Suite 330
 #     Boston, MA 02111-1307
 #     USA
 #
 #   Or contact:
 #
 #    "Thomas Linden" <tom@daemon.de>
 #


#
# this module provides functions used for authentication
# and user/group administration and related stuff
#


package PodWiki::Auth;

$PodWiki::Auth::VERSION = "1.00";



use lib qw(. lib);

use strict;
no strict "refs";

use Exporter;

use Data::Dumper;
use PodWiki::Struct;
use PodWiki::Runner;
use Digest::MD5  qw(md5_hex);

use vars qw(@ISA @EXPORT @EXPORT_OK @EXPORT_TAGS);

@ISA=qw(Exporter);
@EXPORT=qw(fetch_user save_user  create_user_home is_member set_member
           set_members del_member  del_user del_group fetch_user_prefs
           fetch_session check_credentials set_pref_cookie set_cookie
           change_root_passwd access_denied check_user_credentials
           check_auth do_change_passwd);
@EXPORT_OK=qw();
@EXPORT_TAGS=();




sub fetch_user {
  my $user = shift;
  &PodWiki::Runner::tie_users;
  if (exists $PodWiki::Runner::users->{$user}) {
    my($pw, $mail) = split /:/, $PodWiki::Runner::users->{$user};
    return ($user, $pw, $mail);
  }
  else {
    return (0,0,0);
  }
}

sub save_user {
  my ($user, $passwd, $mail) = @_;
  &PodWiki::Runner::tie_users;
  if (! exists $PodWiki::Runner::users->{$user}) {
    &create_user_home($user);
  }
  my(undef, $p, $m) = &fetch_user($user);
  if (! $passwd) {
    $passwd = $p;
  }
  else {
    $passwd = md5_hex($passwd);
  }
  if (! $mail) {
    $mail = $m || "";
  }
  $PodWiki::Runner::users->{$user} = $passwd . ":" . $mail;
  PodWiki::Struct::store($PodWiki::Runner::users, "$PodWiki::Runner::config->{data}/.podusers");
}

sub create_user_home {
  my $user = shift;
  my $home = "UserHome_${user}";
  my $rcs = new PodWiki::Revision(@PodWiki::Runner::rcs, "$home");

  open T, &PodWiki::Runner::untaint(">$PodWiki::Runner::config->{data}/$home")
    or die "Could not write page $home: $!\n";
  print T qq(=begin options

write = user:${user}

=end options

=head2 Home of ${user}

This is your new personal page.

Please L<Edit|state:edit?page=UserHome_${user}> it now.

=cut
  );
  close T;

  if (! -e "$PodWiki::Runner::config->{data}/${home},v") {
    # initial checkin
    $rcs->add('InitialCheckinByPodWiki');
    $rcs->checkout();
  }
  else {
    my $user;
    $PodWiki::Runner::input->{comment} = "*** empty log message ***";
    $PodWiki::Runner::input->{login} = $PodWiki::Runner::current_user || getpwuid($<);
    $rcs->commit("$PodWiki::Runner::input->{comment}", "$PodWiki::Runner::input->{login}");
  }
}

sub is_member {
  my ($user, @groups) = @_;
  &PodWiki::Runner::tie_groups;
  foreach my $group (@groups) {
    if (exists $PodWiki::Runner::groups->{$group}) {
      if (grep {$user eq $_} split /\s*,\s*/, $PodWiki::Runner::groups->{$group} ) {
	# user *IS* a listed member
	return 1;
      }
    }
  }
  # not found
  return 0;
}

sub set_member {
  my($user, $group) = @_;
  &PodWiki::Runner::tie_groups;
  my @users = ();
  if (scalar keys %{$PodWiki::Runner::groups} > 0) {
    # there are groups
    @users = split /\s*,\s*/, $PodWiki::Runner::groups->{$group};
    if (! grep {$user eq $_} @users) {
      $PodWiki::Runner::groups->{$group} = join ",", (@users, $user);
    }
  }
  else {
    # called by setup(), no groups existing
    $PodWiki::Runner::groups->{$group} = $user;
  }
  PodWiki::Struct::store($PodWiki::Runner::groups, "$PodWiki::Runner::config->{data}/.podgroups");
}

sub set_members {
  my($group, @users) = @_;
  &PodWiki::Runner::tie_groups;
  $PodWiki::Runner::groups->{$group} = join ",", (@users);
  PodWiki::Struct::store($PodWiki::Runner::groups, "$PodWiki::Runner::config->{data}/.podgroups");
}

sub del_member {
  my($user, $group) = @_;
  &PodWiki::Runner::tie_groups;
  if ($group eq "*") {
    # remove $user from all groups
    foreach my $group (keys %{$PodWiki::Runner::groups}) {
      if (grep {$user eq $_} split /\s*,\s*/, $PodWiki::Runner::groups->{$group}) {
	my %gusers = map {$_ => 1} split /\s*,\s*/, $PodWiki::Runner::groups->{$group};
	delete $gusers{$user};
	$PodWiki::Runner::groups->{$group} = join ",", keys %gusers;
      }
    }
  }
  else {
    # remove $user from just one group
    my %gusers = map {$_ => 1} split /\s*,\s*/, $PodWiki::Runner::groups->{$group};
    delete $gusers{$user};
    $PodWiki::Runner::groups->{$group} = join ",", keys %gusers;
  }
  PodWiki::Struct::store($PodWiki::Runner::groups, "$PodWiki::Runner::config->{data}/.podgroups");
}

sub del_user {
  my ($user) = @_;
  &PodWiki::Runner::tie_users;
  delete $PodWiki::Runner::users->{$user};
  PodWiki::Struct::store($PodWiki::Runner::users, "$PodWiki::Runner::config->{data}/.podusers");
}

sub del_group {
  my ($group) = @_;
  &PodWiki::Runner::tie_groups;
  delete $PodWiki::Runner::groups->{$group};
  PodWiki::Struct::store($PodWiki::Runner::groups, "$PodWiki::Runner::config->{data}/.podgroups");
}

sub fetch_user_prefs {
  my %cookie = fetch CGI::Cookie;
  delete $cookie{session} if exists $cookie{session};
  foreach my $option (keys %cookie) {
    my $key = $option;
    next if $key eq "session";
    $key =~ s/^user\-//;
    $PodWiki::Runner::input->{$key} = $cookie{$option}->value();
  }
}

sub fetch_session {
  my %cookie = fetch CGI::Cookie;
  if (exists $cookie{session}) {
    # expected cookie exists, user seems to be already logged in
    my ($user, $hashpw) = split /:/, $cookie{session}->value();
    return ($user, $hashpw);
  }
  else {
    # no cookie
    return ("", 0);
  }
}

sub check_user_credentials {
  my ($authuser) = @_;
  my($user, undef) = &fetch_session();
  if ($authuser eq $user) {
    &PodWiki::Runner::errlog("auth", "$authuser == $user");
    return 1;
  }
  else {
    &PodWiki::Runner::errlog("auth", "$authuser != $user");
    return 0;
  }
}

sub check_credentials {
  my @groups = @_;
  my(%session, %cookie, $user, $hashpw, $md5pw, $mail);

  if (exists $PodWiki::Runner::input->{user} && exists $PodWiki::Runner::input->{passwd}) {
    # CGI login, do not use eventually existing cookie
    $user   = $PodWiki::Runner::input->{user};
    $hashpw = md5_hex($PodWiki::Runner::input->{passwd});
  }
  else {
    # expected cookie exists, user seems to be already logged in
    ($user, $hashpw) = &fetch_session;
    if (! $user) {
      # no cookie
      &PodWiki::Runner::errlog("auth", "expected cookie not set");
      return 0;
    }
  }

  # validate credentials(either supplied per cookie or CGI)
  ($user, $md5pw, $mail)  = &fetch_user($user);
  if ($hashpw eq $md5pw) {
    # success, user is authenticated
    &set_cookie($user, $hashpw);
    $PodWiki::Runner::current_user = $user;
    if (@groups) {
      # group info available, check membership
      if (&is_member($user, @groups)) {
	# and is a member in one of the configured groups
	return 1;
      }
      else {
	# no member
	&PodWiki::Runner::errlog("auth", "$user is not a member in @groups");
	return 0;
      }
    }
    else {
      # no group info, membership is not interesting for the moment,
      # grant access now
      return 1;
    }
  }
  else {
    # invalid passwd
    &PodWiki::Runner::errlog("auth", "$user supplied invalid password: $hashpw");
    return 0;
  }
}

sub set_pref_cookie {
  # set user preferences cookie
  my (%prefs) = @_;
  foreach my $option (sort keys %prefs) {
    my $cookie = new CGI::Cookie(
				 -name  => "user-" . $option,
				 -value => $prefs{$option},
				 -expire => "+10y"
				 );
    print "Set-Cookie: $cookie\n";
  }
}

sub set_cookie {
  my($user, $hashpw, $expire) = @_;
  if (! $expire) {
    $expire = $PodWiki::Runner::config->{expire};
  }
  my $cookie = new CGI::Cookie(
			       -name    => "session",
			       -value   => "$user:$hashpw",
			       -expires => "$expire"
			       );
  print "Set-Cookie: $cookie\n";
}

sub change_root_passwd {
  &PodWiki::Runner::change_passwd("root", "Please change the <b>root</b> password!");
}



sub check_auth {
  if (! &PodWiki::Runner::data_empty) {
    my $authuser = "";
    &PodWiki::Runner::parse_options($PodWiki::Runner::page);

    my(@wr_groups, @rd_groups);

    (@wr_groups) = split(/\s*,\s*/, $PodWiki::Runner::page_conf->{write})
      if(exists $PodWiki::Runner::page_conf->{write});

    (@rd_groups) = split(/\s*,\s*/, $PodWiki::Runner::page_conf->{read})
      if(exists $PodWiki::Runner::page_conf->{read});

    if (scalar @wr_groups > 0) {
      if ($wr_groups[0] =~ /user:(.+)/) {
	# user only protection
	@wr_groups = ();
	$authuser  = $1;
      }
    }

    if (scalar @rd_groups > 0) {
      if ($rd_groups[0] =~ /user:(.+)/) {
	# user only protection
	@rd_groups = ();
	$authuser  = $1;
      }
    }

    if ($PodWiki::Runner::sub2go =~ /^edit|save|remove|doremove|checkout$/) {
      if ($PodWiki::Runner::page_conf->{write}) {
	# user wants to edit the page but it is
	# write protected check credentials
	if ($authuser) {
	  if (&check_credentials() && &check_user_credentials($authuser) ) {
	    # granted
	    return;
	  }
	  else {
	    &PodWiki::Runner::errlog("auth", "$authuser ! granted");
	    &access_denied("user write access denied");
	  }
	}
	else {
	  if (! &check_credentials(@wr_groups) ) {
	    &PodWiki::Runner::errlog("auth", "@wr_groups ! granted");
	    &access_denied("group write access denied");
	  }
	}
      }
      else {
	($PodWiki::Runner::current_user, undef) = &fetch_session;
      }
    }
    elsif ($PodWiki::Runner::sub2go eq "check_login") {
      # step in, check supplied credentials and build a cookie
      # session for this user, then call check_auth() again
      # do not attempt to check membership because this info
      # is not available yet
      if (! &check_credentials() ) {
	$PodWiki::Runner::sub2go = $PodWiki::Runner::input->{nextstate};
	&access_denied("login failed");
      }
      else {
	# access granted
	if (! $PodWiki::Runner::input->{nextstate}) {
	  # direct login
	  if (-e "$PodWiki::Runner::config->{data}/UserHome_$PodWiki::Runner::input->{user}") {
	    $PodWiki::Runner::page = "UserHome_$PodWiki::Runner::input->{user}";
	  }
	  $PodWiki::Runner::sub2go = "";
	}
	else {
	  $PodWiki::Runner::sub2go = $PodWiki::Runner::input->{nextstate};
	}
      }
    } # if(check_login)
    elsif ($PodWiki::Runner::sub2go eq "logout") {
      &set_cookie($PodWiki::Runner::current_user, "X", "-d1");
      $PodWiki::Runner::sub2go = "";
    }
    elsif ($PodWiki::Runner::sub2go =~ /_admin$/) {
      # root permissions required!
      if (! &check_credentials("root")) {
	&access_denied("root access denied");
      }
      # root access granted
    }
    else {
      if ($PodWiki::Runner::page_conf->{read}) {
	# page is not publicly readable
	# check credentials
	if ($authuser) {
	  if (&check_credentials() && &check_user_credentials($authuser) ) {
	    # granted
	    return;
	  }
	  else {
	    &PodWiki::Runner::errlog("auth", "$authuser ! granted");
	    &access_denied("user read access denied");
	  }
	}
	else {
	  if (! &check_credentials(@rd_groups) ) {
	    &access_denied("group read access denied");
	  }
	}
      }
      else {
	# unprotected, just deliver the page
	($PodWiki::Runner::current_user, undef) = &fetch_session;
	return;
      }
    }
  } # if(data_empty)
}



sub access_denied {
  # present the login box within normal page content
  my $message    = shift;
  $PodWiki::Runner::input->{nextstate}   = $PodWiki::Runner::sub2go;
  $PodWiki::Runner::input->{nextmessage} = $message;
  $PodWiki::Runner::sub2go        = "login";
  return;
}



sub do_change_passwd {
  my($user, $md5pw, $mail) = &fetch_user($PodWiki::Runner::input->{user});
  if (! $PodWiki::Runner::input->{passwd_old} && $PodWiki::Runner::input->{mail}) {
    # just change the password
    &save_user($user, "", $PodWiki::Runner::input->{mail});
    if (-e "$PodWiki::Runner::config->{data}/UserHome_${$user}") {
      &PodWiki::Runner::render(&PodWiki::Runner::fetch_page("UserHome_${$user}"));
    }
    else {
      &PodWiki::Runner::render(&PodWiki::Runner::fetch_page("PodWikiIndex"));
    }
    print qq(<p class="nolink">Password changed successfully</p>);
  }
  else {
    # change password
    my $hashpw = md5_hex($PodWiki::Runner::input->{passwd_old});
    my $newpw  = md5_hex($PodWiki::Runner::input->{passwd_new});
    if (($md5pw eq $hashpw) && ($PodWiki::Runner::input->{passwd_new} eq $PodWiki::Runner::input->{passwd_repeat})) {
      &save_user($user, $PodWiki::Runner::input->{passwd_new}, $PodWiki::Runner::input->{mail});
      &PodWiki::Runner::render(&PodWiki::Runner::fetch_page("PodWikiIndex"));
      print qq(<p class="nolink">Password changed successfully</p>);

      # FIXME: maybe the user must be logged out here
      # FIXME: unclear: has the user to be logged in to be able to change password?
    }
    else {
      &PodWiki::Runner::change_passwd($user, "Change password - retry");
      print qq(<p class="nolink">Old password invalid or new passwords are not identical</p>);
    }
  }
}



1;

==> podwiki-0.8.3.tgz: podwiki-0.8.3/lib/PodWiki/Auth.pm <==
 #
 #   $Id: Auth.pm,v 1.7 2005/03/24 23:02:33 zarahg Exp $
 #
 #   This  file  is  part of the  PodWiki  web-based authoring tool.
 #
 #   By  accessing  this software,  PodWiki,  you are  duly informed
 #   of and  agree to be  bound  by the  conditions  described below
 #   in this notice:
 #
 #   This software product, PodWiki,  is developed by  Thomas Linden
 #   and   copyrighted  (C) 2003-2005   by  Thomas Linden,  with all
 #   rights reserved.
 #
 #   There is  no charge for  PodWiki software. You can redistribute
 #   it and/or modify it under  the terms of the  GNU General Public
 #   License, which is incorporated by reference herein.
 #
 #   PodWiki is distributed WITHOUT ANY WARRANTY,IMPLIED OR EXPRESS,
 #   OF  MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE or that
 #   the  use of it will not infringe on any third party's intellec-
 #   tual property rights.
 #
 #   You  should  have  received a  copy  of  the GNU General Public
 #   License along with PodWiki.  Copies  can also be obtained from:
 #
 #     http://www.gnu.org/copyleft/gpl.html
 #
 #   or by writing to:
 #
 #     Free Software Foundation, Inc.
 #     59 Temple Place, Suite 330
 #     Boston, MA 02111-1307
 #     USA
 #
 #   Or contact:
 #
 #    "Thomas Linden" <tom@daemon.de>
 #


#
# this module provides functions used for authentication
# and user/group administration and related stuff
#


package PodWiki::Auth;

$PodWiki::Auth::VERSION = "1.02";



use lib qw(. lib);

use strict;
no strict "refs";

use Exporter;

use Data::Dumper;
use PodWiki::Struct;
use PodWiki::Runner;
use Digest::MD5  qw(md5_hex);

use vars qw(@ISA @EXPORT @EXPORT_OK @EXPORT_TAGS);

@ISA=qw(Exporter);
@EXPORT=qw(fetch_user save_user  create_user_home is_member set_member
           set_members del_member  del_user del_group fetch_user_prefs
           fetch_session check_credentials set_pref_cookie set_cookie
           change_root_passwd access_denied check_user_credentials
           check_auth do_change_passwd);
@EXPORT_OK=qw();
@EXPORT_TAGS=();




sub fetch_user {
  my $user = shift;
  &PodWiki::Runner::tie_users;
  if (exists $PodWiki::Runner::users->{$user}) {
    my($pw, $mail) = split /:/, $PodWiki::Runner::users->{$user};
    return ($user, $pw, $mail);
  }
  else {
    return (0,0,0);
  }
}

sub save_user {
  my ($user, $passwd, $mail) = @_;
  &PodWiki::Runner::tie_users;
  if (! exists $PodWiki::Runner::users->{$user}) {
    &create_user_home($user);
  }
  my(undef, $p, $m) = &fetch_user($user);
  if (! $passwd) {
    $passwd = $p;
  }
  else {
    $passwd = md5_hex($passwd);
  }
  if (! $mail) {
    $mail = $m || "";
  }
  $PodWiki::Runner::users->{$user} = $passwd . ":" . $mail;
  PodWiki::Struct::store($PodWiki::Runner::users, "$PodWiki::Runner::config->{data}/.podusers");
}

sub create_user_home {
  my $user = shift;
  my $home = "UserHome_${user}";
  my $rcs = new PodWiki::Revision(@PodWiki::Runner::rcs, "$home");

  open T, &PodWiki::Runner::untaint(">$PodWiki::Runner::config->{data}/$home")
    or die "Could not write page $home: $!\n";
  print T qq(=begin options

write = user:${user}

=end options

=head2 Home of ${user}

This is your new personal page.

Please L<Edit|state:edit?page=UserHome_${user}> it now.

=cut
  );
  close T;

  if (! -e "$PodWiki::Runner::config->{data}/${home},v") {
    # initial checkin
    $rcs->add('InitialCheckinByPodWiki');
    $rcs->checkout();
  }
  else {
    my $user;
    $PodWiki::Runner::input->{comment} = "*** empty log message ***";
    $PodWiki::Runner::input->{login} = $PodWiki::Runner::current_user || getpwuid($<);
    $rcs->commit("$PodWiki::Runner::input->{comment}", "$PodWiki::Runner::input->{login}");
  }
}

sub is_member {
  my ($user, @groups) = @_;
  &PodWiki::Runner::tie_groups;
  foreach my $group (@groups) {
    if (exists $PodWiki::Runner::groups->{$group}) {
      if (grep {$user eq $_} split /\s*,\s*/, $PodWiki::Runner::groups->{$group} ) {
	# user *IS* a listed member
	return 1;
      }
    }
  }
  # not found
  return 0;
}

sub set_member {
  my($user, $group) = @_;
  &PodWiki::Runner::tie_groups;
  my @users = ();
  if (scalar keys %{$PodWiki::Runner::groups} > 0) {
    # there are groups
    @users = split /\s*,\s*/, $PodWiki::Runner::groups->{$group};
    if (! grep {$user eq $_} @users) {
      $PodWiki::Runner::groups->{$group} = join ",", (@users, $user);
    }
  }
  else {
    # called by setup(), no groups existing
    $PodWiki::Runner::groups->{$group} = $user;
  }
  PodWiki::Struct::store($PodWiki::Runner::groups, "$PodWiki::Runner::config->{data}/.podgroups");
}

sub set_members {
  my($group, @users) = @_;
  &PodWiki::Runner::tie_groups;
  $PodWiki::Runner::groups->{$group} = join ",", (@users);
  PodWiki::Struct::store($PodWiki::Runner::groups, "$PodWiki::Runner::config->{data}/.podgroups");
}

sub del_member {
  my($user, $group) = @_;
  &PodWiki::Runner::tie_groups;
  if ($group eq "*") {
    # remove $user from all groups
    foreach my $group (keys %{$PodWiki::Runner::groups}) {
      if (grep {$user eq $_} split /\s*,\s*/, $PodWiki::Runner::groups->{$group}) {
	my %gusers = map {$_ => 1} split /\s*,\s*/, $PodWiki::Runner::groups->{$group};
	delete $gusers{$user};
	$PodWiki::Runner::groups->{$group} = join ",", keys %gusers;
      }
    }
  }
  else {
    # remove $user from just one group
    my %gusers = map {$_ => 1} split /\s*,\s*/, $PodWiki::Runner::groups->{$group};
    delete $gusers{$user};
    $PodWiki::Runner::groups->{$group} = join ",", keys %gusers;
  }
  PodWiki::Struct::store($PodWiki::Runner::groups, "$PodWiki::Runner::config->{data}/.podgroups");
}

sub del_user {
  my ($user) = @_;
  &PodWiki::Runner::tie_users;
  delete $PodWiki::Runner::users->{$user};
  PodWiki::Struct::store($PodWiki::Runner::users, "$PodWiki::Runner::config->{data}/.podusers");
}

sub del_group {
  my ($group) = @_;
  &PodWiki::Runner::tie_groups;
  delete $PodWiki::Runner::groups->{$group};
  PodWiki::Struct::store($PodWiki::Runner::groups, "$PodWiki::Runner::config->{data}/.podgroups");
}

sub fetch_user_prefs {
  my %cookie = fetch CGI::Cookie;
  delete $cookie{session} if exists $cookie{session};
  foreach my $option (keys %cookie) {
    my $key = $option;
    next if $key eq "session";
    $key =~ s/^user\-//;
    $PodWiki::Runner::input->{$key} = $cookie{$option}->value();
  }
}

sub fetch_session {
  my %cookie = fetch CGI::Cookie;
  if (exists $cookie{session}) {
    # expected cookie exists, user seems to be already logged in
    my ($user, $hashpw) = split /:/, $cookie{session}->value();
    return ($user, $hashpw);
  }
  else {
    # no cookie
    return ("", 0);
  }
}

sub check_user_credentials {
  my ($authuser) = @_;
  my($user, undef) = &fetch_session();
  if ($user eq $authuser || $user eq "root") {
    # root overrides
    &PodWiki::Runner::errlog("auth", "$authuser == $user");
    return 1;
  }
  else {
    &PodWiki::Runner::errlog("auth", "$authuser != $user");
    return 0;
  }
}

sub check_credentials {
  my @groups = @_;
  my(%session, %cookie, $user, $hashpw, $md5pw, $mail);

  if (exists $PodWiki::Runner::input->{user} && exists $PodWiki::Runner::input->{passwd}) {
    # CGI login, do not use eventually existing cookie
    $user   = $PodWiki::Runner::input->{user};
    $hashpw = md5_hex($PodWiki::Runner::input->{passwd});
  }
  else {
    # expected cookie exists, user seems to be already logged in
    ($user, $hashpw) = &fetch_session;
    if (! $user) {
      # no cookie
      &PodWiki::Runner::errlog("auth", "expected cookie not set");
      return 0;
    }
  }

  # validate credentials(either supplied per cookie or CGI)
  ($user, $md5pw, $mail)  = &fetch_user($user);
  if ($hashpw eq $md5pw) {
    # success, user is authenticated
    &set_cookie($user, $hashpw);
    $PodWiki::Runner::current_user = $user;
    if (@groups) {
      # group info available, check membership
      if (&is_member($user, @groups)) {
	# and is a member in one of the configured groups
	return 1;
      }
      else {
	# no member
	&PodWiki::Runner::errlog("auth", "$user is not a member in @groups");
	return 0;
      }
    }
    else {
      # no group info, membership is not interesting for the moment,
      # grant access now
      return 1;
    }
  }
  else {
    # invalid passwd
    &PodWiki::Runner::errlog("auth", "$user supplied invalid password: $hashpw");
    return 0;
  }
}

sub set_pref_cookie {
  # set user preferences cookie
  my (%prefs) = @_;
  foreach my $option (sort keys %prefs) {
    my $cookie = new CGI::Cookie(
				 -name  => "user-" . $option,
				 -value => $prefs{$option},
				 -expire => "+10y"
				 );
    print "Set-Cookie: $cookie\n";
  }
}

sub set_cookie {
  my($user, $hashpw, $expire) = @_;
  if (! $expire) {
    $expire = $PodWiki::Runner::config->{expire};
  }
  my $cookie = new CGI::Cookie(
			       -name    => "session",
			       -value   => "$user:$hashpw",
			       -expires => "$expire"
			       );
  print "Set-Cookie: $cookie\n";
}

sub change_root_passwd {
  &PodWiki::Runner::change_passwd("root", "Please change the <b>root</b> password!");
}



sub check_auth {
  if (! &PodWiki::Runner::data_empty) {
    my $authuser = "";
    &PodWiki::Runner::parse_options($PodWiki::Runner::page);

    my(@wr_groups, @rd_groups);

    (@wr_groups) = split(/\s*,\s*/, $PodWiki::Runner::page_conf->{write})
      if(exists $PodWiki::Runner::page_conf->{write});

    (@rd_groups) = split(/\s*,\s*/, $PodWiki::Runner::page_conf->{read})
      if(exists $PodWiki::Runner::page_conf->{read});

    if (scalar @wr_groups > 0) {
      if ($wr_groups[0] =~ /user:(.+)/) {
	# user only protection
	@wr_groups = ();
	$authuser  = $1;
      }
    }

    if (scalar @rd_groups > 0) {
      if ($rd_groups[0] =~ /user:(.+)/) {
	# user only protection
	@rd_groups = ();
	$authuser  = $1;
      }
    }

    if ($PodWiki::Runner::sub2go =~ /^edit|save|remove|doremove|checkout$/) {
      if ($PodWiki::Runner::page_conf->{write}) {
	# user wants to edit the page but it is
	# write protected check credentials
	if ($authuser) {
	  if (&check_credentials() && &check_user_credentials($authuser) ) {
	    # granted
	    return;
	  }
	  else {
	    &PodWiki::Runner::errlog("auth", "$authuser ! granted");
	    &access_denied("user write access denied");
	  }
	}
	else {
	  if (! &check_credentials(@wr_groups) ) {
	    &PodWiki::Runner::errlog("auth", "@wr_groups ! granted");
	    &access_denied("group write access denied");
	  }
	}
      }
      else {
	($PodWiki::Runner::current_user, undef) = &fetch_session;
      }
    }
    elsif ($PodWiki::Runner::sub2go eq "check_login") {
      # step in, check supplied credentials and build a cookie
      # session for this user, then call check_auth() again
      # do not attempt to check membership because this info
      # is not available yet
      if (! &check_credentials() ) {
	$PodWiki::Runner::sub2go = $PodWiki::Runner::input->{nextstate};
	&access_denied("login failed");
      }
      else {
	# access granted
	if (! $PodWiki::Runner::input->{nextstate}) {
	  # direct login
	  if (-e "$PodWiki::Runner::config->{data}/UserHome_$PodWiki::Runner::input->{user}") {
	    $PodWiki::Runner::page = "UserHome_$PodWiki::Runner::input->{user}";
	  }
	  $PodWiki::Runner::sub2go = "";
	}
	else {
	  $PodWiki::Runner::sub2go = $PodWiki::Runner::input->{nextstate};
	}
      }
    } # if(check_login)
    elsif ($PodWiki::Runner::sub2go eq "logout") {
      &set_cookie($PodWiki::Runner::current_user, "X", "-d1");
      $PodWiki::Runner::sub2go = "";
    }
    elsif ($PodWiki::Runner::sub2go =~ /_admin$/) {
      # root permissions required!
      if (! &check_credentials("root")) {
	&access_denied("root access denied");
      }
      # root access granted
    }
    else {
      if ($PodWiki::Runner::page_conf->{read}) {
	# page is not publicly readable
	# check credentials
	if ($authuser) {
	  if (&check_credentials() && &check_user_credentials($authuser) ) {
	    # granted
	    return;
	  }
	  else {
	    &PodWiki::Runner::errlog("auth", "$authuser ! granted");
	    &access_denied("user read access denied");
	  }
	}
	else {
	  if (! &check_credentials(@rd_groups) ) {
	    &access_denied("group read access denied");
	  }
	}
      }
      else {
	# unprotected, just deliver the page
	($PodWiki::Runner::current_user, undef) = &fetch_session;
	return;
      }
    }
  } # if(data_empty)
}



sub access_denied {
  # present the login box within normal page content
  my $message    = shift;
  $PodWiki::Runner::input->{nextstate}   = $PodWiki::Runner::sub2go;
  $PodWiki::Runner::input->{nextmessage} = $message;
  $PodWiki::Runner::sub2go        = "login";
  return;
}



sub do_change_passwd {
  my($user, $md5pw, $mail) = &fetch_user($PodWiki::Runner::input->{changeuser});
  if (! $PodWiki::Runner::input->{passwd_old} && $PodWiki::Runner::input->{mail}) {
    # just change the password
    &save_user($user, "", $PodWiki::Runner::input->{mail});
    if (-e "$PodWiki::Runner::config->{data}/UserHome_${$user}") {
      &PodWiki::Runner::render(&PodWiki::Runner::fetch_page("UserHome_${$user}"));
    }
    else {
      &PodWiki::Runner::render(&PodWiki::Runner::fetch_page("PodWikiIndex"));
    }
    print qq(<p class="nolink">Password changed successfully</p>);
  }
  else {
    # change password
    my $hashpw = md5_hex($PodWiki::Runner::input->{passwd_old});
    my $newpw  = md5_hex($PodWiki::Runner::input->{passwd_new});
    if (($md5pw eq $hashpw) && ($PodWiki::Runner::input->{passwd_new} eq $PodWiki::Runner::input->{passwd_repeat})) {
      &save_user($user, $PodWiki::Runner::input->{passwd_new}, $PodWiki::Runner::input->{mail});
      &PodWiki::Runner::render(&PodWiki::Runner::fetch_page("PodWikiIndex"));
      print qq(<p class="nolink">Password changed successfully</p>);

      # FIXME: maybe the user must be logged out here
      # FIXME: unclear: has the user to be logged in to be able to change password?
    }
    else {
      &PodWiki::Runner::change_passwd($user, "Change password - retry");
      print qq(<p class="nolink">Old password invalid or new passwords are not identical</p>);
    }
  }
}



1;
